import numpy as np


__np_types = {
    'complex128': np.complex128, 'complex256': np.complex256,
    'complex64': np.complex64,
    'float128': np.float128, 'float64': np.float64,
    'float16': np.float16, 'float32': np.float32,
    'int8': np.int8, 'int16': np.int16, 'int32': np.int32, 'int64': np.int64
}


__top_level_sub_xds = '__attrs_xds'

