__active_mask = 'active_mask'
__doppler_types = ['RADIO', 'Z', 'RATIO', 'BETA', 'GAMMA']
__native_types = ['FREQ', 'VRAD', 'VOPT', 'BETA', 'WAVE', 'AWAV']
__object_name = 'object_name'
__pointing_center = 'pointing_center'

